import signal

import click

from camonline.config import CONFIG_PATH, ConfigLoader
from camonline.log import logger
from camonline.monitor import RotateMonitor


@click.command()
@click.option("--config", default=CONFIG_PATH, help=f"Config file path, default: {CONFIG_PATH}")
def start(config):
    config = ConfigLoader(config).load_config()
    m = RotateMonitor(config)
    m.start()

    def _shutdown(sig=None, frame=None):
        logger.info("Exiting...")
        m.shutdown()
        m.camera.shutdown()
        exit(0)

    signal.signal(signal.SIGINT, _shutdown)
    signal.signal(signal.SIGTERM, _shutdown)
    logger.info("Waiting for KeyboardInterrupt or SIGTERM...")
    while True:
        signal.pause()


@click.group()
def cli():
    pass


cli.add_command(start)


if __name__ == "__main__":
    cli(["start"])



import copy
import os
import shutil
from pathlib import Path
from typing import Any, Dict, Optional, Union

import tomli
import tomli_w

from camonline.exceptions import ConfigFileNotFoundError
from camonline.log import logger

_HERE = Path(__file__).parent
DEFAULT_CONFIG = _HERE / "static" / "config.toml"
CONFIG_PATH = "~/.camonline/config.toml"


class Config:
    """
    A wrapper for config dict
    """

    def __init__(self, config_dict: Optional[Dict[str, Any]] = None):
        if not config_dict:
            config_dict = {}
        self._config_dict: Dict[str, Any] = config_dict

    def __repr__(self) -> str:
        return str(self._config_dict)

    def __getattr__(self, name):
        # All config keys are lower case
        name = name.lower()
        if isinstance(self._config_dict.get(name), dict):
            return Config(self._config_dict[name])

        return self._config_dict.get(name, None)

    def __bool__(self):
        return bool(self._config_dict)


class ConfigLoader:
    """
    A loader for config file and environment variables

    User should use CLI for this
    """

    ENV_PREFIX = "CAMONLINE_"
    ENV_SEP = "__"
    DUMP_DIR = "/tmp"

    def __init__(
        self,
        path: Optional[Union[str, Path]] = None,
        load_env: bool = True,
        dump_when_load=True,
        config_dump_dir=None,
        generate_config=True,
    ):
        if not path:
            path = Path(CONFIG_PATH).expanduser()

        self.config_path: Path = Path(path).expanduser().absolute()
        if generate_config and not self.config_path.exists():
            self.generate_config()

        self.load_env = load_env
        self.dump_when_load = dump_when_load
        self.config_dump_dir = config_dump_dir or self.DUMP_DIR

    def __repr__(self) -> str:
        return f"ConfigLoader(path={self.config_path}, dump_when_load={self.dump_when_load}, load_env={self.load_env}, config_dump_dir={self.config_dump_dir})"

    def _init_default_modules(self, config_dict: Dict[str, Any]) -> Dict[str, Any]:
        config_dict.setdefault("tracer", {})
        config_dict.setdefault("collector", {})
        config_dict.setdefault("filter", {})
        config_dict.setdefault("monitor", {})
        return config_dict

    def generate_config(self):
        logger.info(f"Creating default config file {self.config_path}")
        self.config_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(DEFAULT_CONFIG, self.config_path)

    def normalize_config(self, config_dict: Dict[str, Any]) -> Dict[str, Any]:
        # Make sure all config keys are lower case
        for k in list(config_dict.keys()):
            v = config_dict[k]
            if isinstance(v, dict):
                config_dict[k] = self.normalize_config(v)
            if k.lower() != k:
                config_dict[k.lower()] = config_dict.pop(k)

        return config_dict

    def load_config(self) -> Dict[str, Any]:
        logger.info(f"Loading config from {self.config_path}")
        if not self.config_path.exists():
            raise ConfigFileNotFoundError(f"Config file:{self.config_path} not found.")

        try:
            with self.config_path.open("rb") as f:
                config = tomli.load(f)

                config = self._init_default_modules(config)
                if self.load_env:
                    config = self.load_env_config(config)

                config = self.normalize_config(config)

                if self.dump_when_load:
                    # Dump current config to a tmp file
                    config_dump_path = (
                        Path(self.config_dump_dir) / f"camonline_config.toml.{os.getpid()}"
                    )
                    self.dump_config(config, config_dump_path)
                return config
        except tomli.TOMLDecodeError as e:
            logger.error(f"Error loading config: {e}")
            raise e

    def load_env_config(self, config_dict: Dict[str, Any]) -> Dict[str, Any]:
        logger.info(
            f"Loading config from environment variables, prefix: `{self.ENV_PREFIX}`, sep: `{self.ENV_SEP}`"
        )
        for k, v in os.environ.items():
            if not k.startswith(self.ENV_PREFIX):
                continue
            k = k[len(self.ENV_PREFIX) :]
            k = k.lower()
            logger.debug(f"Loading {k.replace(self.ENV_SEP, '.')}={v}")
            *index, spec = k.split(self.ENV_SEP)
            last = config_dict
            for i in index:
                last = last.setdefault(i, {})
            last[spec] = v
        return config_dict

    def dump_config(self, config_dict: Dict[str, Any], path: Union[str, Path]):
        dump_path = Path(path).expanduser().resolve()
        dump_path.parent.mkdir(parents=True, exist_ok=True)
        with dump_path.open("wb") as f:
            tomli_w.dump(config_dict, f)
        logger.info(f"Current config has been dumped to {dump_path}")


class Configuable:
    """
    A base class for all configuable classes

    default_config: Dict[str, Any]  default config for this class
    config_scope: str
        config scope for this class, e.g. tracer, collector
    """

    default_config = {}
    config_scope: Optional[str] = None

    def __init__(self, config: Optional[Union[Config, Dict[str, Any]]] = None, *args, **kwargs):
        if not config:
            config = {}
        elif isinstance(config, Config):
            config = config._config_dict

        if self.config_scope:
            for score in self.config_scope.split("."):
                config = config.get(score.lower(), {})
        c = copy.deepcopy(self.default_config)

        def _recursive_update(c, config):
            for k, v in config.items():
                if not isinstance(v, dict):
                    c[k] = v
                else:
                    c.setdefault(k, {})
                    _recursive_update(c[k], v)

        _recursive_update(c, config)
        self.config = Config(c)
        logger.debug(f"{self.__class__.__name__} config loaded.")

    def __repr__(self):
        return f"{self.__class__.__name__}({self.config})"



import threading

import cv2 as cv


class CameraManager:
    _cache = {}
    _mutex = threading.Lock()

    @classmethod
    def get_or_create(cls, index):
        if index not in cls._cache:
            with cls._mutex:
                if index not in cls._cache:
                    cls._cache[index] = Camera(index)
        return cls._cache[index]


class Camera:
    def __init__(self, index) -> None:
        self.cap = cv.VideoCapture(index)
        self.callbacks = []

    def _read_fream(self):
        ret, frame = self.cap.read()
        if not ret:
            return None
        return frame

    def add_callback(self, callback):
        self.callbacks.append(callback)

    def remove_callback(self, callback):
        if callback in self.callbacks:
            self.callbacks.remove(callback)

    def poll(self):
        frame = self._read_fream()
        if frame is None:
            return
        for callback in self.callbacks:
            callback(frame)

    def shutdown(self):
        self.cap.release()


if __name__ == "__main__":
    index = 4
    c = Camera(index)

    def show_frame(frame):
        cv.imshow("frame", frame)
        cv.waitKey(1)

    c.add_callback(show_frame)

    while True:
        c.poll()



import os

os.chdir(os.path.dirname(os.path.abspath(__file__)))
os.environ["CAMONLINE_LOG_LEVEL"] = "DEBUG"

import importlib.metadata as importlib_metadata
import re
import sys
from pathlib import Path


def load_entry_point(distribution, group, name):
    dist_obj = importlib_metadata.distribution(distribution)
    eps = [ep for ep in dist_obj.entry_points if ep.group == group and ep.name == name]
    if not eps:
        raise ImportError("Entry point %r not found" % ((group, name),))
    return eps[0].load()


config_file = Path("./config.toml")

if __name__ == "__main__":
    sys.argv[0] = re.sub(r"(-script\.pyw?|\.exe)?$", "", sys.argv[0])
    sys.argv.append("start")
    sys.argv.extend(["--config", config_file.resolve().as_posix()])
    sys.exit(load_entry_point("camonline", "console_scripts", "camonline")())



import threading
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Dict, Optional, Union

import cv2 as cv

from camonline.camera import Camera, CameraManager
from camonline.config import Config, Configuable
from camonline.log import logger


class RotateMonitor(Configuable):
    default_config = {
        "camera": {
            "device": 0,
            "resolution": [640, 480],
            "fps": 30.0,
        },
        "storage": {
            "record_dir": "~/.camonline/storage",
            "fourcc": "PIM1",
            "suffix": ".mkv",
            "days": 30,
        },
    }

    def __init__(self, config: Optional[Union[Config, Dict[str, Any]]] = None, *args, **kwargs):
        super().__init__(config, *args, **kwargs)
        self.camera: Camera = CameraManager.get_or_create(self.config.camera.device)
        self.record_dir.mkdir(parents=True, exist_ok=True)
        self.start_day = datetime.now().strftime("%Y-%m-%d")
        self.shutdown_event = threading.Event()

        self._attatch_func = None
        self._shutdown_callback = None
        self._thread = None
        self._rotate_thread = None

    @property
    def record_dir(self) -> Path:
        return Path(self.config.storage.record_dir).expanduser().absolute()

    @property
    def current_record_file(self) -> Path:
        today = datetime.now().strftime("%Y-%m-%d")
        return (self.record_dir / today).with_suffix(self.config.storage.suffix).as_posix()

    def attatch(self):
        fourcc = cv.VideoWriter_fourcc(*self.config.storage.fourcc)
        out = cv.VideoWriter(
            self.current_record_file,
            fourcc,
            self.config.camera.fps,
            self.config.camera.resolution,
        )

        def _(frame):
            logger.debug(frame)
            out.write(frame)

        logger.info(f"Attatch to {self.current_record_file}")
        self._attatch_func = _
        self.camera.add_callback(_)

        def _close():
            out.release()

        self._shutdown_callback = _close

    def reattatch(self):
        logger.info("Reattatch triggerd")
        if self._attatch_func:
            self.camera.remove_callback(self._attatch_func)
        if self._shutdown_callback:
            self._shutdown_callback()
        self.attatch()

    def start(self):
        self.shutdown_event.clear()
        self.attatch()

        def _():
            while not self.shutdown_event.is_set():
                self.camera.poll()
                self.shutdown_event.wait(1 / self.config.camera.fps)

        def _rotate():
            while not self.shutdown_event.is_set():
                today = datetime.now().strftime("%Y-%m-%d")
                if today != self.start_day:
                    self.reattatch()
                    self.start_day = today
                    to_remove_day = (
                        datetime.now() - timedelta(days=self.config.storage.days)
                    ).strftime("%Y-%m-%d")
                    logger.info(
                        f"Delete {self.config.storage.days} days before record:{to_remove_day}"
                    )
                    (self.record_dir / to_remove_day).with_suffix(
                        self.config.storage.suffix
                    ).unlink(missing_ok=True)

                self.shutdown_event.wait(1)

        self._thread = threading.Thread(target=_)
        self._thread.start()

        self._rotate_thread = threading.Thread(target=_rotate)
        self._rotate_thread.start()

    def shutdown(self):
        self.shutdown_event.set()
        self._thread.join()
        self._rotate_thread.join()

        self.camera.remove_callback(self._attatch_func)

        self._shutdown_callback()
        self._shutdown_callback = None

        self._attatch_func = None
        self._thread = None
        self._rotate_thread = None


if __name__ == "__main__":
    from pathlib import Path

    from camonline.config import ConfigLoader

    _HERE = Path(__file__).parent
    config_loader = ConfigLoader(_HERE / "static" / "config.toml")

    monitor = RotateMonitor(config_loader.load_config())
    monitor.start()
    input()
    monitor.shutdown()
